"""
This module provides the remote workspace
"""

from typing import Optional, Callable, Any, Iterable, cast
from json import dumps
from inspect import signature
from textwrap import dedent
from dfiibridge_core import DfiiBridgeClient, Session, LoggingCallback

from .hints import SkillCode, Skill, Function, Symbol
from .functions import FunctionCollection, RemoteFunction
from .translator import Translator
from .logger import SimpleStdoutLogger
from .abstract_workspace import AbstractWorkspace

def _python_value_to_skill(value: Skill, is_on_top_level:bool=True) -> SkillCode:
    """
    When "is_on_top_level" is True, we inject a call to __pyReconstructSkillObjects,
    so that SKILL evaluates the symbols starting from the top level.
    For all nested/deeper elements this variable is false to continue with regular processing.
    """
    try:
        return value.__repr_skill__()  # type: ignore
    except AttributeError:
        pass

    if isinstance(value, dict):
        items = ' '.join(f"{key} {_python_value_to_skill(value, is_on_top_level=False)}" for key, value in value.items())
        if not is_on_top_level:
            return SkillCode(f'(nil {items})')
        return SkillCode(f'__pyReconstructSkillObjects(\'(nil {items}))')

    if value is False or value is None:
        return SkillCode('nil')

    if value is True:
        return SkillCode('t')

    if isinstance(value, (int, float, str)):
        return SkillCode(dumps(value))

    if isinstance(value, (list, tuple)):
        inner = ' '.join(_python_value_to_skill(item, is_on_top_level=False) for item in value)
        if not is_on_top_level:
            return SkillCode(f'({inner})')
        return SkillCode(f'__pyReconstructSkillObjects(\'({inner}))')

    type_ = type(value).__name__
    raise RuntimeError(f"Cannot convert object {type_!r} to skill.") from None

def encode_call(func_name: str, *args: Skill, **kwargs: Skill) -> SkillCode:
    args_code = ' '.join(map(_python_value_to_skill, args))
    kw_keys = kwargs
    kw_values = map(_python_value_to_skill, kwargs.values())
    kwargs_code = ' '.join(f'?{key} {value}' for key, value in zip(kw_keys, kw_values))
    return SkillCode(f'{func_name}({args_code} {kwargs_code})')

class Workspace(AbstractWorkspace):

    def __init__(self, session: Session):
        self.session = session
        self._client = DfiiBridgeClient()
        self._client.connect_to_endpoint(session) # connect to remote session

        self._translator = Translator(self.execute_raw)
        self.timeout = -1
        # add function collection "skill" to the workspace
        # usage: self.skill.functionName(...)
        setattr(self, "skill", FunctionCollection(self, self._translator))

    def set_origin_session(self, session: Session):
        """Register the origin session so that you can call Python function from
        skill via DfiiBridgeClient_executeCommand().

        Args:
            session: Session of the server that runs on the client side
        """
        self._client.set_origin_session(session)

    def set_timeout(self, timeout:int):
        """
        Set the timeout (in seconds) of the tcp connection between client and Virtuoso server.
        Timeout = -1 removes the timeout
        """
        self.timeout = timeout

    def clear_timeout(self):
        """Set the timeout to -1 and therefore disable it"""
        self.timeout = -1

    def get_virtuoso_verbosity(self) -> int:
        """ Get the verbosity of the virtuoso server.
        """
        return int(self._client.execute("get_server_loglevel", [])[0])

    def set_virtuoso_verbosity(self, verbosity:int):
        """ Set the verbosity of the virtuoso server.
        More messages lead to slower speed because of logging overhead.

        10: Critical + Error + Warning + Debug + Info messages
        20: Critical + Error + Warning + Debug        messages
        30: Critical + Error + Warning                messages
        40: Critical + Error                          messages
        50: Critical                                  messages
        """
        self._client.execute("set_server_loglevel", [str(verbosity)])

    def enable_debug_mode(self):
        """
        Enable verbose mode for client and server
        """
        # Increase verbosity on the server side
        self.set_virtuoso_verbosity(10)
        self._simple_logger = SimpleStdoutLogger()
        self.register_logger(self._simple_logger)

    def register_logger(self, logger:LoggingCallback):
        """
        Register a logger for the client (must be an instance of the subclass of LoggingCallback).
        """
        self._client.register_logging_function(logger)

    def __getitem__(self, item: str) -> RemoteFunction:
        """
        The workspace can be used like a dictionary to return a function object.
        """
        return RemoteFunction(self, item, self._translator)

    def execute(self, func_name: str, *args, **kwargs) -> any:
        """
        Executes a remote SKILL function with arguments 'args' and returns the result
        """
        return self.__getitem__(func_name)(*args, **kwargs)

    def execute_raw(self, data:str, parse_return_value:bool=False, request_reply:bool=True, timeout:int=None) -> any:

        if timeout is None:
            timeout = self.timeout

        if parse_return_value and not request_reply:
            raise Exception("You can only parse the return value if you are also receiving it!")

        # There should be one or none response
        ret = self._client.execute("skill", [data], request_reply, timeout)

        if request_reply:
            parsed_ret = self.parse_response(ret[0]) # Evaluate success/fail message
            if parse_return_value:
                return self._translator.decode(parsed_ret)
            return parsed_ret.strip() # Strip whitespace (\n at end of line)
        return None # Return nothing if no return is expected

    def set(self, var_name:str, value:any) -> None:
        """
        Define a variable on the skill side
        """
        cmd = f"{var_name} = {self._translator.encode(value)}"
        self.execute_raw(data=cmd, request_reply=True)

    def get(self, var_name:str) -> any:
        """
        Get a variable on the skill side
        """
        return self.execute_raw(data=var_name, parse_return_value=True, request_reply=True)

    def define(self, name: str, args: Iterable[str], code: str) -> None:
        """
        Define a function on the skill side.
        """
        code = code.replace('\n', ' ')
        arg_list = ' '.join(args)
        code = f'defun(user{name} ({arg_list}) {code})'
        cast(Symbol, self._translator.decode(self.execute_raw(code)))

    @staticmethod
    def parse_response(response: str) -> str:
        """
        Parse the string response from virtuoso.
        It contains the status (True/False), which is stripped from the response.
        """
        if response.startswith("failure "):
            _, err_msg = response.split(' ', maxsplit=1)
            raise RuntimeError(err_msg)

        return response

    @staticmethod
    def fix_completion() -> None:
        try:
            ipy = get_ipython()  # type: ignore
        except NameError:
            pass
        else:
            ipy.Completer.use_jedi = False
            ipy.Completer.greedy = True

    @staticmethod
    def _build_function(function: Callable[..., Any]) -> Function:
        if not function.__doc__:
            raise RuntimeError("Function does not have a doc string.")

        s = signature(function)

        if s.return_annotation is s.empty:
            raise RuntimeError("Function does not have a return annotation.")

        param_doc = []
        for p in s.parameters.values():
            if p.default is p.empty:
                param = p.name

                if p.annotation is Optional:
                    param = f"    [ {param} ]"
                else:
                    param = f"    {param}"
            else:
                param = f"    [ ?{p.default} {p.name} ]"

            param_doc.append(param)

        doc = [
            function.__name__ + "(",
            *param_doc,
            f"=> {'nil' if s.return_annotation is None else s.return_annotation}",
            "",
        ]

        doc_string = '\n'.join(doc) + dedent(function.__doc__)

        return Function(function.__name__, doc_string, set())



from abc import ABC, abstractmethod
from typing import Callable
from .hints import Skill, SkillCode

class ParseError(Exception):
    pass

class AbstractTranslator(ABC):

    def __init__(self, execute_raw : Callable) -> None:
        self.execute_raw = execute_raw

    @abstractmethod
    def encode_call(self, func_name: str, *args: Skill, **kwargs: Skill) -> SkillCode:
        return NotImplemented

    @abstractmethod
    def encode_dir(self, obj: SkillCode) -> SkillCode:
        return NotImplemented

    @abstractmethod
    def decode_dir(self, code: str) -> list[str]:
        return NotImplemented

    @abstractmethod
    def encode_getattr(self, obj: SkillCode, key: str) -> SkillCode:
        return NotImplemented

    @abstractmethod
    def encode_globals(self, filter_regex: str) -> SkillCode:
        return NotImplemented

    @abstractmethod
    def decode_globals(self, code: str) -> list[str]:
        return NotImplemented

    @abstractmethod
    def encode_help(self, symbol: str) -> SkillCode:
        return NotImplemented

    @abstractmethod
    def decode_help(self, help_: str) -> str:
        return NotImplemented

    @abstractmethod
    def encode_setattr(self, obj: SkillCode, key: str, value: any) -> SkillCode:
        return NotImplemented

    @abstractmethod
    def encode(self, value: Skill) -> SkillCode:
        return NotImplemented

    @abstractmethod
    def decode(self, code: str) -> Skill:
        return NotImplemented


from abc import ABC, abstractmethod
from .hints import Skill, SkillCode

class AbstractWorkspace(ABC):

    @abstractmethod
    def execute_raw(self, data:str, parse_return_value:bool=False, request_reply:bool=True, timeout:int=-1) -> any:
        return NotImplemented


#!/usr/bin/env python3

"""
Handles the calls from SKILL:
- read origin, search_criteria, function and function arguments from stdin
- parse them and send them to each custom server that matches the criteria given

Stdin contains the following:
    srcHost:srcPort
    search_criteria
    function_name
    [arguments]\n * N
    <<<<END>>>>

"""

import sys
from dfiibridge_core import DfiiBridgeClient, CoreSessionManager, LoggingCallback
from argparse import ArgumentParser
import os
import io


class ClientLoggingCallback(LoggingCallback):

    def __init__(self, file_name, verbosity:int=0):
        super().__init__()

        self.file_name = file_name
        self.verbosity = verbosity

    def log(self, log_level, log_message):
        #with open(self.file_name,'a') as f:
        #    f.write(f"{datetime.datetime.now()} {log_level}: {log_message}\n")
        print(log_message)

def main(args):

    # read in all lines including the line break from stdin, until <END> marker hit
    raw_args = []

    while True:
        cur_line = sys.stdin.readline()
        if cur_line.strip() == "<<<<END>>>>":
            break
        raw_args += [cur_line.strip()]

    # find sessions for the criteria provided in the message
    csm = CoreSessionManager(args.session_dir)
    client = DfiiBridgeClient(csm)


    if args.hostname and args.port:
        sessions = [csm.get_session(args.hostname, args.port, False)]
    elif args.search_criteria:
        # extract userdata search from args.search_criteria
        userdata = {}
        if args.search_criteria:
            for var in args.search_criteria.split(","):
                k,v = var.split("=")
                userdata[k] = v
        sessions = list(csm.find_sessions(userdata))
    else:
        raise Exception("You need to define a search criteria or a hostname/port!")

    if not sessions:
        raise Exception(f"No sessions are currently running with the filter {userdata}!")

    origin_session = csm.get_session(args.origin_hostname, args.origin_port, False)

    for session in sessions:
        if session == origin_session:
            sessions.remove(session)
            break

    if not sessions:
        raise Exception(f"No session found with the userdata {userdata} (only one but this is the origin session!)")

    if not sessions:
        raise Exception("No session is left. Maybe you try to connect to the origin session? "
                        "This is not possible because it would end up in a self blocking loop.")

    client.set_origin_session(origin_session)

    # check if there is a session with these criteria available
    if not sessions:
        print(f"no sessions found for identifier '{userdata}'!")
        sys.exit(-1)

    # Enable logger if wished
    if args.verbosity<10:
        logger = ClientLoggingCallback("dfiibridge_logger.log")
        client.register_logging_function(logger)

    client.connect_to_endpoints( sessions )

    replies = client.execute(args.function_name, raw_args)

    for reply in replies:
        print("\u2764"+reply, file=sys.stderr, end="") # give reponse for each client

if __name__ == '__main__':

    argument_parser = ArgumentParser(sys.argv[0])
    argument_parser.add_argument('--session_dir',
                                type=str,
                                default=os.path.join(os.environ["HOME"], ".dfiiBridge"),
                                help='Directory to store the session file (default: ~/.dfiiBridge)')
    argument_parser.add_argument('--origin_hostname',
                                type=str,
                                required=True,
                                help='Hostname of the virtuoso_server')
    argument_parser.add_argument('--origin_port',
                                type=int,
                                required=True,
                                help='Port of the virtuoso_server')
    argument_parser.add_argument('--function_name',
                                type=str,
                                required=True,
                                help='Name of the function to call remotely')
    argument_parser.add_argument('-v', '--verbosity', default=30, type=int, help="Set the veribosity level")
    argument_parser.add_argument('--search_criteria',
                                type=str,
                                default="",
                                help="Search criteria ('key1=value1,key2=value2')")
    argument_parser.add_argument('--hostname',
                                type=str,
                                default=None,
                                help="Define the hostname of your target")
    argument_parser.add_argument('--port',
                                type=int,
                                default=None,
                                help="Define the port of your target")
    args = argument_parser.parse_args()

    try:
        main(args)
    except Exception as e:
        print(f"\u2764error(\"{e}\")", file=sys.stderr) # give reponse for each client



"""
Python API for executing commands in a remote virtuoso session

- allows to connect to virtuoso session running on remote machine
- manages virtuoso sessions in ~/.dfiiBridge
- adds user authentication to command execution to prevent unauthorized users
  from using the command interface

The most relevant objects are:

- sessions.VirtuosoSessionManager to list sessions and to connect to them
- workspace.Workspace to execute skill commands in the remote session
"""

from .workspace import Workspace
from .sessions import VirtuosoSessionManager, Session

from .logger import configure_logger, LOGGER_FILE_NAME, LOGGER_NAME
from .callback_decorator import callback, skill_callback, log_callback



import time
import json
import subprocess
import os
import re
import sys
from select import select
from typing import Callable

from .globals import DFII_LABEL_PATTERN, COMMAND_VIRTUOSO, DFII_SERVER_PY, DFII_SERVER_SKILL
class TimeoutException(Exception):
    pass
class DfiiBridgeServerNotStartedException(Exception):
    pass

def data_available_with_timeout(stream, timeout):
    """
    This is to check if output is available,
    before actually calling read/readline,
    so that non-blocking reads can be achieved.
    """
    readable, _, _ = select([stream], [], [], timeout)

    return bool(readable)
class SessionMaker:
    """
    Class that provides methods for session creation.
    """

    @staticmethod
    def create_session(no_cdsinit:bool=False,
                       no_gui:bool=False,
                       cmd:Callable=None,
                       env:dict=None,
                       session_root:str=None,
                       extra_fields:dict=None,
                       python_executable:str=None):
        """
        Start a Virtuoso session, inside or outside of Camino (depending on using_camino).
        """

        # build virtuoso command string
        init = "-nocdsinit" if no_cdsinit else ""
        gui = "-nograph" if no_gui else ""

        if cmd is None:
            cmd = COMMAND_VIRTUOSO

        command_virtuoso_f = cmd(
            init=init,
            gui=gui,
            path_to_skill_server=DFII_SERVER_SKILL
        )

        if env is None:
            env = os.environ.copy()  # Use current environment
        if extra_fields is None:
            extra_fields = {}
        if python_executable is None:
            python_executable = sys.executable

        # Update path to the executable
        env["DFIIBRIDGE_DFII_SERVER"] = python_executable + " " + DFII_SERVER_PY
        env["DFIIBRIDGE_EXTRA"] = json.dumps(extra_fields)
        if session_root is not None:
            env["DFIIBRIDGE_SESSION_ROOT"] = session_root

        # start the process and return the created process
        print(f"Starting: {command_virtuoso_f}")

        proc = subprocess.Popen(
            command_virtuoso_f,
            stdout=subprocess.PIPE,
            stdin=subprocess.PIPE, # we also need stdin because otherwise the normal stdin is messed up
            stderr=subprocess.STDOUT,
            env=env
        )
        return proc

    @staticmethod
    def _wait_for_file(caller_self, session_manager, session_adress_status, on_output):
        # look for the session file so we can read it
        session_file = session_manager.get_session_file_path(session_adress_status['hostname'],
                                                             int(session_adress_status['port']),
                                                             False)

        # wait until session file becomes alive
        if on_output:
            on_output(caller_self, "Waiting until sessionfile available...\n")

        w = 0
        while os.path.exists(session_file) == False:
            time.sleep(1)
            w += 1

            if w > 30:  # longer than 30sec => error
                raise TimeoutException("Timeout while waiting for sessionfile...")

        # the session file is available, so we can now get the real session object
        return session_manager.get_session(session_file, False)

    @staticmethod
    def wait_for_session(proc, session_manager, caller_self=None, on_output=None):
        """
        Wait for the session to start. The process was already created.
        """
        # wait until the process completes
        while True:
            # check if the process is still alive
            if proc.poll() is not None:
                break

            # check if data is available on process' stdout
            if data_available_with_timeout(proc.stdout, 240):

                # read from stdout now that output is available
                out = proc.stdout.readline().decode("utf-8")

                # try to find the session creation pattern
                m = re.findall(DFII_LABEL_PATTERN, out)

                if m:
                    # convert it into dictionary and emits it as new created session.
                    # Python server sends string beginning with '%%%DfiiBridge%%%' label
                    # followed by a dictionary.
                    session_adress_status = {
                        "hostname": m[0][0],
                        "port": m[0][1],
                        "status": m[0][2]
                    }

                    # get session data on success
                    if session_adress_status["status"] == "OK":

                        # notify about success
                        if on_output:
                            on_output(caller_self, "Session Status OK\n")

                        # return session object
                        return SessionMaker._wait_for_file(caller_self, session_manager, session_adress_status, on_output)
                    else:
                        # status was not OK, return None (failure)
                        raise DfiiBridgeServerNotStartedException(f"Status of server: {session_adress_status['status']}")

                else:
                    # call output handler, if available
                    if on_output:
                        on_output(caller_self, out)

            else:
                # call stdout wait timeout handler, if available
                raise TimeoutException("Timeout while waiting for output...")

        # If the session was not started successfully (wherever we exited the while loop),
        # we kill the session manually.
        raise DfiiBridgeServerNotStartedException("DfiiBridgeServer and/or Virtuoso could not be started!")



"""
This module provides object translation between SKILL and Python

Internal package - does not need to be used
"""

from typing import NoReturn, Any, List, Iterable, cast
from json import dumps, loads
from warnings import warn_explicit

from .hints import SkillCode, Skill, Symbol
from .abstract_translator import AbstractTranslator
from .abstract_workspace import AbstractWorkspace
from .objects import RemoteObject, RemoteTable, RemoteVector


class Translator(AbstractTranslator):

    def _raise_error(message: str) -> NoReturn:
        raise ParseError(message)

    def _show_warning(self, message: str, result: Any) -> Any:
        for i, line in enumerate(message.splitlines(keepends=False)):
            warn_explicit(line.lstrip("*WARNING*"), UserWarning, "Skill response", i)

        return result

    def _skill_value_to_python(self, string: str) -> Skill:
        return eval(  # type: ignore
            string,
            {
                'Remote': lambda skill_string: RemoteObject(skill_string, self),
                'Table': lambda skill_string: RemoteTable(skill_string, self),
                'Vector': lambda skill_string: RemoteVector(skill_string, self),
                'Symbol': Symbol,
                'error': self._raise_error,
                'warning': self._show_warning
            }
        )

    def _python_value_to_skill(self, value: Skill, is_on_top_level:bool=True) -> SkillCode:
        """
        When "is_on_top_level" is True, we inject a call to DfiiBridge__pyReconstructSkillObjects,
        so that SKILL evaluates the symbols starting from the top level.
        For all nested/deeper elements this variable is false to continue with regular processing.
        """
        try:
            return value.__repr_skill__()  # type: ignore
        except AttributeError:
            pass

        if isinstance(value, dict):
            items = ' '.join(f"{key} {self._python_value_to_skill(value, is_on_top_level=False)}" for key, value in value.items())
            if not is_on_top_level:
                return SkillCode(f'(nil {items})')
            return SkillCode(f'DfiiBridge__pyReconstructSkillObjects(\'(nil {items}))')

        if value is False or value is None:
            return SkillCode('nil')

        if value is True:
            return SkillCode('t')

        if isinstance(value, (int, float, str)):
            return SkillCode(dumps(value))

        if isinstance(value, (list, tuple)):
            inner = ' '.join(self._python_value_to_skill(item, is_on_top_level=False) for item in value)
            if not is_on_top_level:
                return SkillCode(f'({inner})')
            return SkillCode(f'DfiiBridge__pyReconstructSkillObjects(\'({inner}))')

        type_ = type(value).__name__
        raise RuntimeError(f"Cannot convert object {type_!r} to skill.") from None

    def build_skill_path(self, components: Iterable[str]) -> SkillCode:
        it = iter(components)
        path = str(next(it))

        for component in it:
            if isinstance(component, int):
                path = f'(nth {component} {path})'
            else:
                path = f'{path}->{component}'

        return SkillCode(path)


    def build_python_path(self, components: Iterable[str]) -> SkillCode:
        it = iter(components)
        path = str(next(it))

        for component in it:
            if isinstance(component, int):
                path = f'{path}[{component}]'
            else:
                path = f'{path}.{component}'

        return SkillCode(path)


    def encode_call(self, func_name: str, *args: Skill, **kwargs: Skill) -> SkillCode:
        args_code = ' '.join(map(self._python_value_to_skill, args))
        kw_keys = kwargs
        kw_values = map(self._python_value_to_skill, kwargs.values())
        kwargs_code = ' '.join(f'?{key} {value}' for key, value in zip(kw_keys, kw_values))
        return SkillCode(f'{func_name}({args_code} {kwargs_code})')

    def encode_dir(self, obj: SkillCode) -> SkillCode:
        parts = ' '.join(
            (
                f'{obj}->?',
                f"if( type({obj}) == 'rodObj then {obj}->systemHandleNames)",
                f'if( type({obj}) == \'rodObj then {obj}->userHandleNames)',
            )
        )
        code = f'mapcar(lambda((attr) sprintf(nil "%s" attr)) nconc({parts}))'
        return SkillCode(code)

    def decode_dir(self, code: str) -> List[str]:
        attributes = self._skill_value_to_python(code) or ()
        return [attr for attr in cast(List[str], attributes)]

    def encode_getattr(self, obj: SkillCode, key: str) -> SkillCode:
        return self.build_skill_path([obj, key])

    def encode_globals(self, filter_regex: str) -> SkillCode:
        return SkillCode(f'buildString(listFunctions("{filter_regex}"))')

    def decode_globals(self, code: str) -> List[str]:
        return [f for f in loads(code).split()]

    def encode_help(self, symbol: str) -> SkillCode:
        code = f"""
            _text = outstring()
            poport = _text help({symbol})
            poport = stdout getOutstring(_text)
        """.replace(
            "\n", " "
        )
        return SkillCode(code)

    def decode_help(self, help_: str) -> str:
        return loads(help_)  # type: ignore

    def encode_setattr(self, obj: SkillCode, key: str, value: Any) -> SkillCode:
        code = self.build_skill_path([obj, key])
        value = self._python_value_to_skill(value)
        return SkillCode(f'{code} = {value}')

    def encode(self, value: Skill) -> SkillCode:
        return self._python_value_to_skill(value)

    def decode(self, code: str) -> Skill:
        return self._skill_value_to_python(code)




import re

import importlib.resources

COMMAND_VIRTUOSO = lambda gui, init, path_to_skill_server: ["virtuoso", gui, init, "-replay", path_to_skill_server]

# Session creation pattern to find DfiiBridge information in Virtuoso stdout
DFII_LABEL_PATTERN = r"^\[INFO\]\s+DfiiBridge\s+running\s+on\s+(.*?):(\d+)\s+status:(\w+)"

DFII_LABEL_PATTERN = re.compile(DFII_LABEL_PATTERN, re.VERBOSE)

DFII_CLIENT_SKILL = str(importlib.resources.files(__package__) / "dfiiBridgeClient.il")
DFII_SERVER_SKILL = str(importlib.resources.files(__package__) / "dfiiBridgeServer.il")
DFII_UTILS_SKILL =  str(importlib.resources.files(__package__) / "dfiiBridgeUtils.il")
DFII_SERVER_PY =    str(importlib.resources.files(__package__) / "virtuoso_server.py")
DFII_CLIENT_PY =    str(importlib.resources.files(__package__) / "virtuoso_client.py")



#!/usr/bin/env python3
"""
Python server that receives requests and, if they are valid, executes them in virtuoso.
"""

import sys
from sys import stdout, stdin, stderr
import os
import re
from argparse import ArgumentParser
import subprocess
import json
import datetime
import glob

from dfiibridge_server.helpers import execute_inside_skill
from dfiibridge import VirtuosoSessionManager
from dfiibridge_core import DfiiBridgeServer, RemoteCallback, LoggingCallback, CoreSessionManager
from dfiibridge import callback, log_callback
import dfiibridge_server

import logging
import signal
import sys


global log_file_name, log_level
log_file_name: str = None
log_level:int = 30

@callback
def get_server_loglevel(ws) -> str:
    """ Get the loglevel
    """
    global log_level
    return str(log_level)

@callback
def set_server_loglevel(ws, _log_level:str) -> str:
    """ Set the loglevel
    """
    global log_level
    log_level = int(_log_level)
    execute_inside_skill(f"DfiiBridge['server]['verbosity] = {log_level}")
    write_to_log(f"[INFO] Setting dfiibridge server loglevel to {log_level}")
    return "True"

@log_callback
def log(log_level_in, log_message):
    """ Logging callback (see https://docs.python.org/3/library/logging.html#logging-levels)
        10: Critical + Error + Warning + Debug + Info messages
        20: Critical + Error + Warning + Debug        messages
        30: Critical + Error + Warning                messages
        40: Critical + Error                          messages
        50: Critical                                  messages
    """
    global log_level

    if log_level > log_level_in:
        return

    write_to_log(f"{datetime.datetime.now()} {log_level_in}: {log_message}\n")

@callback
def skill(ws, skill_code:str) -> str:
    return execute_inside_skill(skill_code)


def write_to_log(msg:str, mode:str="a"):
    global log_file_name

    with open(log_file_name, mode) as f:
        f.write(msg.strip()+"\n")

def cleanup_log(csm : CoreSessionManager):
    """Clean up the session directory of old log files

    Args:
        csm: Session Manager
    """
    root = csm.get_root_directory()
    logs = list(glob.glob(os.path.join(root, "*.log")))
    # Create a list with (file name, modification time stamp)
    files = [(log_file, os.stat(log_file).st_mtime) for log_file in logs]
    # Sort for modification time stamp (newest one is at 0)
    files.sort(key=lambda entry: entry[1], reverse=True)

    # Do not allow more than 10 log files -> remove oldest ones
    if len(files)>10:
        for file in files[-10:]:
            try:
                os.remove(file[0])
            except Exception:
                continue # jump of non-deletable files

class VirtuosoServer:

    def __init__(self, session_dir:str, verbosity:int, user_data:str, cds_log_path:str):
        """ VirtuosoServer
        Extend this class and define our custom userdata structure for your design environment.
        """
        global log_file_name, log_level

        log_level = int(verbosity)

        csm = CoreSessionManager(session_dir)

        cleanup_log(csm)

        # Additional extra fields can be passed by the environment variable $DFIIBRIDGE_EXTRA
        # It can be set in the session_maker.py create_session() function call
        extra_fields = json.loads(user_data)

        # Add the cds log path
        extra_fields.update({"cds_log": cds_log_path})

        self.server = DfiiBridgeServer(csm, {
                "type": "virtuoso",
                **extra_fields
            })

        # Register the skill callback handler
        log_file_name = csm.get_log_file_path(self.server.get_hostname(), self.server.get_port())

        write_to_log(f"""
##################################################################################################
########################################### DfiiBridge ###########################################
##################################################################################################
    Server started:          {datetime.datetime.now()}
    Hostname:                {self.server.get_hostname()}
    Port:                    {self.server.get_port()}
    Log file:                {log_file_name}
    Session directory:       {session_dir}
    Session file:            {self.server.get_session_file_path()}
    Verbosity:               {verbosity}
    DfiiBridge-Tool-Version: {dfiibridge_server.__version__}
##################################################################################################""", mode="w")

        self.server.register_function("skill", skill)
        self.server.register_function("set_server_loglevel", set_server_loglevel)
        self.server.register_function("get_server_loglevel", get_server_loglevel)
        self.server.register_logging_function(log)

        # Save session information for skill
        execute_inside_skill(f"DfiiBridge['server]['log] = \"{log_file_name}\"")
        execute_inside_skill(f"DfiiBridge['server]['hostname] = \"{self.server.get_hostname()}\"")
        execute_inside_skill(f"DfiiBridge['server]['port] = {self.server.get_port()}")
        execute_inside_skill(f"DfiiBridge['server]['session_file] = \"{self.server.get_session_file_path()}\"")
        execute_inside_skill(f"setShellEnvVar(\"DFIIBRIDGE_VIRTUOSO_SESSION_FILE\" \"{self.server.get_session_file_path()}\")")

        # Echo for Session Manager GUI
        execute_inside_skill("""system("echo [INFO] DfiiBridge running on """ +
                            f"""{self.server.get_hostname()}:{self.server.get_port()} status:OK")""")

        sys.stdout.flush()

    def listen(self):
        """Start listening
        """
        try:
            self.server.listen() # blocking
        except RuntimeError as e:
            print(f"[ERROR] DfiiBridge encountered an error: {e}")
            write_to_log(f"[ERROR] DfiiBridge encountered an error: {e}")
            raise
        except Exception as e:
            write_to_log(f"[INFO] {e}. This exception is okay, if the server is being terminated by the user.")

        server.stop()

        print("[INFO] DfiiBridge terminates now.")

    def stop(self):
        self.server.unregister_logging_function(log)
        self.server.stop()
        write_to_log("[INFO] Server stopped.")

if __name__ == '__main__':

    argument_parser = ArgumentParser(sys.argv[0])
    argument_parser.add_argument('--session_dir',
                                type=str,
                                default=os.path.join(os.environ["HOME"], ".dfiiBridge"),
                                help='Directory to store the session file (default: ~/.dfiiBridge)')
    argument_parser.add_argument('--user_data',
                                type=str,
                                default="{}",
                                help='User data that is attached to the session, e.g. project information (default: {})')
    argument_parser.add_argument('--cdslog',
                                type=str,
                                default="",
                                help='Path to the cdslog file of the current session (default: "")')

    argument_parser.add_argument('-v', '--verbose', default=30, help="Set the veribosity level")

    args = argument_parser.parse_args()

    server = VirtuosoServer(args.session_dir, args.verbose, args.user_data, args.cdslog)
    server.listen()





from dfiibridge_core import DfiiBridgeClient, RemoteCallback, LoggingCallback
from .workspace import Workspace

def log_callback(func):
    """Decorator for LoggingCallbacks

    Args:
        func (Callable): Function that looks like: def my_func(log_level, log_message)->None

    Returns:
        LoggingCallback: Instance of the LoggingCallback
    """
    class DecoratorLoggingCallback(LoggingCallback):
        def log(self, log_level, log_message):
            try:
                func(log_level, log_message)
            except Exception as e:
                print(f"ERROR: {e}")
                raise
    return DecoratorLoggingCallback()

def callback(func):
    """Decorator for RemoteCallbacks

    Args:
        func (Callable): Function that looks like: def my_func(origin_client:DfiiBridgeClient, raw_args:list)->str

    Returns:
        RemoteCallback: Instance of the RemoteCallback
    """
    class DecoratorCallback(RemoteCallback):
        def callback(self, origin_client:DfiiBridgeClient, raw_args:list)->str:
            try:
                return func(origin_client, *raw_args)
            except Exception as e:
                return f"failure {e}"
    return DecoratorCallback()


def skill_callback(func):
    """Decorator for RemoteCallbacks that can be called from skill

    This decorator does the translation of the arguments and the translation of the return type
    for you.

    Args:
        func (Callable): Function that looks like: def my_func(origin_ws:Workspace, args:list)->str

    Returns:
        RemoteCallback: Instance of the RemoteCallback
    """
    class DecoratorCallback(RemoteCallback):
        def callback(self, origin_client:DfiiBridgeClient, raw_args:list)->str:

            try:
                if not origin_client:
                    return "failure No origin server found! Is it really listening?"

                origin_sessions = origin_client.get_connected_sessions()

                if len(origin_sessions)!=1:
                    return "failure No or multiple connected sessions on origin_client! This should not happen."

                origin_ws = Workspace(origin_sessions[0])

                func_args = []
                for arg in raw_args:
                    func_args.append(origin_ws._translator.decode(arg))

                answ = func(origin_ws, *func_args)
                return origin_ws._translator.encode(answ)

            except Exception as e:
                return f"failure {e}"

    return DecoratorCallback()



"""
This module provides SKILL function abstractions and
multiple functions (collections) for remote execution.

Internal package - does not need to be used
"""

from typing import List
from functools import lru_cache

from .hints import SkillCode, Skill, Key
from .abstract_translator import AbstractTranslator

class FunctionCollection:
    """
    Dictionary of remote function objects. This collection contains all functions.
    Contents are looked up in Virtuoso and cached for later usage
    """
    def __init__(self, ws, translator: AbstractTranslator):
        self._workspace = ws
        self._translate = translator

    def __repr__(self) -> str:
        """
        Print the list of functions nicely
        """
        return f"<function collection*>\n{dir(self)}"

    @lru_cache(maxsize=128)
    def __dir__(self) -> List[str]:
        """
        Create command to query list of functions and execute it, afterwards cache the results.
        """
        command = self._translate.encode_globals("^[A-Za-z]")  # all functions
        result = self._workspace.execute_raw(command, parse_return_value=False)
        return self._translate.decode_globals(result)

    def __getattr__(self, item: str) -> 'RemoteFunction':
        """
        Looking up the function in the dictionary returns the function object
        """
        return RemoteFunction(self._workspace, item, self._translate)


class RemoteFunction:
    """
    Function object for function that can be executed in Virtuoso
    """
    def __init__(self, ws, func: str, translator: AbstractTranslator):
        self._workspace = ws
        self._translate = translator
        self._function = func

    def __call__(self, *args: Skill, **kwargs: Skill) -> Skill:
        """
        Call function in Virtuoso using the channel and return the result
        """
        command = self.lazy(*args, **kwargs)
        return self._workspace.execute_raw(command, parse_return_value=True)

    def lazy(self, *args: Skill, **kwargs: Skill) -> SkillCode:
        """
        Create function call for Virtuoso, but do not call it
        """
        return self._translate.encode_call(self._function, *args, **kwargs)

    def __repr__(self) -> str:
        """
        Print the function help instead of calling it
        """
        command = self._translate.encode_help(self._function)
        result = self._workspace.execute_raw(command, parse_return_value=False)
        return self._translate.decode_help(result)



#!/usr/bin/env python
# -*- coding:utf-8 -*-

"""Setup this SWIG library."""
import os
import re
import shutil
import subprocess

from setuptools import Extension, setup
from setuptools.command.build_py import build_py
from setuptools.command.install import install
from setuptools.command.egg_info import egg_info

STD_EXT = Extension(
    name='_dfiibridge_core',
    swig_opts=['-c++', '-Iswig/include', '-threads', '-py3'],
    sources=[
        'swig/src/DfiiBridgeServer.cpp',
        'swig/src/Channel.cpp',
        'swig/src/ChannelClient.cpp',
        'swig/src/ChannelServer.cpp',
        'swig/src/CoreSessionManager.cpp',
        'swig/src/DfiiBridgeClient.cpp',
        'swig/src/DfiiBridgeCore.cpp',
        'swig/src/Session.cpp',
        'swig/src/DfiiBridge.i'
    ],
    include_dirs=[
        'swig/include',
    ],
    extra_compile_args=['-std=c++17'],
    libraries = ['stdc++fs', 'rt', 'pthread']
)



"""
Build extensions before python modules, or the generated SWIG python files will be missing.
"""
class BuildPy(build_py):

    def setup_libzmq(self):
        """
        Add libzmq to the list of needed link libraries
        """

        libs_stdout = subprocess.run(["gcc", "-lzmq"], stderr=subprocess.PIPE).stderr

        if "cannot find -lzmq" in libs_stdout.decode():
            raise Exception("Libzmq is not present on your system. Please install it.")

        print("Libzmq is already installed. I'll link to it.")
        STD_EXT.libraries += ["zmq"]

    def verify_swig_version(self):
        """
        Verify that the required SSIG version 4 is available.
        """
        stdout_swig_version = subprocess.check_output(["swig", "-version"])
        swig_version = re.findall(r"(\d+)\.(\d+)\.(\d+)", stdout_swig_version.decode())
        if swig_version:
            if swig_version[0][0]!="4":
                raise Exception(f"The swig version {swig_version} that is installed is too old")
                # we need a new version of swig!


    def run(self):
        self.setup_libzmq()
        self.verify_swig_version()
        self.run_command('build_ext')
        super(build_py, self).run()


class PostInstallCommand(install):
    """
    We need to copy the SWIG generated python file DfiiBridgeCore.py in the installation directory.
    This file copy isn't done by the framework. Without this python script, none of our python classes
    are available.
    """

    def run(self):
        print(f"Copy the SWIG generated python file to {self.install_libbase}")
        install.run(self)
        shutil.copyfile("swig/src/dfiibridge_core.py", os.path.join(self.install_libbase, "dfiibridge_core.py"))

class egg_info_ex(egg_info):
    """
    Includes license file into `.egg-info` folder.
    SRC: https://stackoverflow.com/questions/9977889/how-to-include-license-file-in-setup-py-script
    """

    def run(self):
        # don't duplicate license into `.egg-info` when building a distribution
        if not self.distribution.have_run.get('install', True):
            # `install` command is in progress, copy license
            self.mkpath(self.egg_info)
            self.copy_file('LICENSE', self.egg_info)

        egg_info.run(self)

setup(
    ext_modules=[STD_EXT],
    cmdclass={'build_py': BuildPy , 'install': PostInstallCommand, 'egg_info': egg_info_ex},
    license_files = ('LICENSE',),
    license = 'GPL3',
)



"""
This module provides functions to communicate between the server and virtuoso
"""

from sys import stdin, stderr


def execute_inside_skill(data: str) -> str:
    """
    Send skill code to Virtuoso and receive the reply
    """
    stderr.write(">"+data.replace("\n", "\u2764"))
    stderr.write("\n")
    stderr.flush()
    return stdin.readline()





"""
This module provides client and server logging - console and file based

Internal package - does not need to be used
"""

import logging
from dfiibridge_core import LoggingCallback

LOGGER_NAME = "dfii_bridge"
LOGGER_FILE_NAME = "dfii_bridge.log"


class CustomFormatter(logging.Formatter):
    """Logging Formatter to add colors and count warning / errors
    SRC: https://stackoverflow.com/questions/384076/how-can-i-color-python-logging-output
    """

    green = "\033[32;1m"
    yellow = "\033[33;1m"
    red = "\033[31;1m"
    reset = "\033[0m"

    FORMATS = {
        logging.DEBUG: green + "[%(levelname)s]" + reset + " %(asctime)s - %(name)s - %(message)s (%(filename)s:%(lineno)d)",
        logging.INFO: green + "[%(levelname)s]" + reset + " %(asctime)s - %(name)s - %(message)s (%(filename)s:%(lineno)d)",
        logging.WARNING: yellow + "[%(levelname)s]" + reset + " %(asctime)s - %(name)s - %(message)s (%(filename)s:%(lineno)d)",
        logging.ERROR: red + "[%(levelname)s]" + reset + " %(asctime)s - %(name)s - %(message)s (%(filename)s:%(lineno)d)",
        logging.CRITICAL: red + "[%(levelname)s]" + reset + " %(asctime)s - %(name)s - %(message)s (%(filename)s:%(lineno)d)",
    }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)


def configure_logger(only_file=False, log_level=logging.INFO) -> None:
    """
    Create and initialize the logger, by setting the log level and adding
    multiple handlers for console and file output
    """
    logger = logging.getLogger(LOGGER_NAME)
    logger.setLevel(log_level)

    if not only_file:
        ch = logging.StreamHandler()
        ch.setLevel(log_level)
        ch.setFormatter(CustomFormatter())
        logger.addHandler(ch)

    fh = logging.FileHandler(LOGGER_FILE_NAME)
    fh.setLevel(log_level)
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    fh.setFormatter(formatter)
    logger.addHandler(fh)

class SimpleStdoutLogger(LoggingCallback):
    """
    Simple implementation of the logging interface LoggingCallback
    """

    def __init__(self, log_level=logging.DEBUG):
        super().__init__()
        self.log_level = log_level

    def log(self, log_level, log_message):
        '''Log the passed message if it matches my log level
           The log levels are descending defined: CRITICAL > ERROR > WARING > INFO > DEBUG
        '''
        if log_level >= self.log_level:
            print(log_message)



"""
This module provides object mapping between SKILL and Python

Internal package - does not need to be used
"""

from typing import Any, List, Optional, cast, Union, overload, Sequence, Iterable, MutableMapping, Iterator, Callable

from .functions import RemoteFunction
from .hints import SkillCode, Symbol, Var, Skill
from .abstract_translator import AbstractTranslator, ParseError

def is_jupyter_magic(attribute: str) -> bool:
    ignore = {
        '_ipython_canary_method_should_not_exist_',
        '_ipython_display_',
        '_repr_mimebundle_',
        '_repr_html_',
        '_repr_markdown_',
        '_repr_svg_',
        '_repr_png_',
        '_repr_pdf_',
        '_repr_jpeg_',
        '_repr_latex_',
        '_repr_json_',
        '_repr_javascript_',
        '_rapped',
        '__wrapped__',
        '__call__',
    }
    return attribute in ignore


class RemoteVariable:
    _attributes = {'_execute_raw', '_variable', '_translate'}

    def __init__(self, variable: str, translator: AbstractTranslator) -> None:
        self._execute_raw = translator.execute_raw
        self._variable = SkillCode(variable)  # example "__py_db_0x2259de1a"
        self._translate = translator

    def __repr_skill__(self) -> SkillCode:
        return SkillCode(self._variable)

    def _send(self, command: SkillCode) -> Any:
        return self._execute_raw(command).strip()

    def __repr__(self) -> str:
        return self.__str__()

    def _call(self, function: str, *args: Skill, **kwargs: Skill) -> Skill:
        code = self._translate.encode_call(function, *args, **kwargs)
        result = self._send(code)
        return self._translate.decode(result)

    def __dir__(self) -> Iterable[str]:
        response = self._send(self._translate.encode_dir(self._variable))
        attributes = self._translate.decode_dir(response)
        return attributes

class RemoteObject(RemoteVariable):

    @property
    def skill_id(self) -> int:
        addr = self._variable[5:].rsplit('_', maxsplit=1)[1]  # "__py_db_0x2259de1a" -> "0x2259de1a"
        return addr

    @property
    def skill_parent_type(self) -> str:
        return self._variable[5:].rsplit('_', maxsplit=1)[0]

    def _is_open_file(self) -> bool:
        return self._variable.startswith('__py_openfile_')

    @property
    def skill_type(self) -> Optional[str]:
        if self._is_open_file():
            return 'open_file'

        try:
            typ = self.obj_type
        except RuntimeError:
            return None
        if typ is None:
            return None
        if isinstance(typ, Symbol):
            return typ.name[2:-4]
        return cast(str, typ)

    def __str__(self) -> str:
        typ = self.skill_type or self.skill_parent_type
        if typ == 'open_file':
            result = self._send(self._translate.encode_call('sprintf', None, '%s', self))
            name = self._translate.decode(result)[6:-1]  # type: ignore
            return f"<remote open_file {name!r}>"
        return f"<remote {typ}@{self.skill_id}>"

    def __dir__(self) -> Iterable[str]:
        if self._is_open_file():
            return super().__dir__()

        response = self._send(self._translate.encode_dir(self._variable))
        attributes = self._translate.decode_dir(response)
        return attributes

    def __getattr__(self, key: str) -> Any:
        if is_jupyter_magic(key):
            raise AttributeError(key)

        result = self._send(self._translate.encode_getattr(self._variable, key))
        return self._translate.decode(result)

    def __setattr__(self, key: str, value: Any) -> None:
        if key in RemoteObject._attributes:
            return super().__setattr__(key, value)

        result = self._send(self._translate.encode_setattr(self._variable, key, value))
        self._translate.decode(result)

    def getdoc(self) -> str:
        return "Properties:\n- " + '\n- '.join(dir(self))

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, RemoteObject):
            return self._variable == other._variable
        return NotImplemented

    def __ne__(self, other: Any) -> bool:
        if isinstance(other, RemoteObject):
            return self._variable != other._variable
        return NotImplemented

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        command = self._translate.encode_call('funcall', self, *args, **kwargs)
        result = self._send(code)
        return self._translate.decode(result)

    @property
    def lazy(self) -> 'LazyList':
        return LazyList(self._variable, self._translate)


class LazyList(RemoteVariable):
    arg = Var('arg')

    def __getattr__(self, attribute: str) -> 'LazyList':
        variable = SkillCode(f"{self._variable}~>{attribute}")
        return LazyList(variable, self._translator)

    def __str__(self) -> str:
        return f"<lazy list {self._variable}>"

    @staticmethod
    def _condition(filters: Sequence[SkillCode]) -> SkillCode:
        if len(filters) == 1:
            return SkillCode(f'arg->{filters[0]}')
        parameters = ' '.join(f'arg->{f}' for f in filters)
        return SkillCode(f'and({parameters})')

    def filter(self, *args: str, **kwargs: Any) -> 'LazyList':
        if not args and not kwargs:
            return self

        arg_filters = [SkillCode(arg) for arg in args]
        kwarg_filters = [
            SkillCode(f"{key} == {self._translator.encode(value)}")
            for key, value in kwargs.items()
        ]
        filters = self._condition(arg_filters + kwarg_filters)
        variable = SkillCode(f'setof(arg {self._variable} {filters})')

        return LazyList(variable, self._translator)

    @overload
    def __getitem__(self, item: int) -> RemoteObject:
        ...  # pragma: nocover

    @overload  # noqa
    def __getitem__(self, item: slice) -> List[RemoteObject]:  # noqa
        ...  # pragma: nocover

    def __getitem__(self, item: Union[int, slice]) -> Union[RemoteObject, List[RemoteObject]]:  # noqa
        if isinstance(item, int):
            code = self._translator.encode_call('nth', item, Var(self._variable))
        else:
            if item.start is not None or item.stop is not None or item.step is not None:
                raise RuntimeError("cannot slice lazy list with arbitrary bounds")

            code = self._variable

        result = self._send(code)
        return self._translator.decode(result)  # type: ignore

    def __len__(self) -> int:
        code = self._translator.encode_call('length', Var(self._variable))
        return self._translator.decode(self._send(code))  # type: ignore

    def foreach(self, func: Union['RemoteFunction', SkillCode], *args: Any) -> None:
        if isinstance(func, RemoteFunction):
            args = args or (LazyList.arg,)
            func = func.lazy(*args)
        elif args:
            raise RuntimeError("cannot combine args with remote function")

        code = self._translator.encode_call('foreach', LazyList.arg, Var(self._variable), Var(func))
        result = self._send(code + ',nil')
        self._translator.decode(result)



class RemoteCollection(RemoteVariable):
    def __str__(self) -> str:
        return f'<remote {self._call("lsprintf", "%L", self)}>'

    def __len__(self) -> int:
        return cast(int, self._call('length', self))

    def __getitem__(self, item: Skill) -> Skill:
        return self._call('arrayref', self, item)

    def __setitem__(self, key: Skill, value: Skill) -> None:
        self._call('setarray', self, key, value)

    def __delitem__(self, item: Skill) -> None:
        self._call('remove', item, self)

class RemoteTable(RemoteCollection, MutableMapping[Skill, Skill]):
    def __getitem__(self, item: Skill) -> Skill:
        try:
            return super().__getitem__(item)
        except ParseError:
            raise KeyError(item) from None

    def __getattr__(self, item: str) -> Skill:
        return self[Symbol(item)]

    def __setattr__(self, key: str, value: Skill) -> None:
        if key in self._attributes:
            super().__setattr__(key, value)
        else:
            self[Symbol(key)] = value

    def __iter__(self) -> Iterator[Skill]:
        code = self._translate.encode_getattr(self.__repr_skill__(), '?')
        result = self._send(code)
        return iter(self._translate.decode(result) or ())  # type: ignore


class RemoteVector(RemoteCollection):
    def __getitem__(self, item: Skill) -> Skill:
        try:
            return super().__getitem__(item)
        except RuntimeError as e:
            if "array index out of bounds" in str(e):
                raise IndexError(f"list index {item} out of range (len={len(self)})") from None
            raise  # pragma: no cover
        except ParseError:
            raise IndexError(f"list index {item} out of range (len={len(self)})") from None

    def __setitem__(self, item: Skill, value: Skill) -> None:
        try:
            super().__setitem__(item, value)
        except RuntimeError as e:
            if "array index out of bounds" in str(e):
                raise IndexError(f"list index {item} out of range (len={len(self)})") from None
            raise  # pragma: no cover


import logging
import os
from typing import Callable, List

from dfiibridge_core import CoreSessionManager, Session

from .workspace import Workspace
from .logger import LOGGER_NAME

from dfiibridge_server.session_maker import SessionMaker


logger = logging.getLogger(LOGGER_NAME)


class VirtuosoSessionManager(CoreSessionManager):

    def get_workspace(self, session: Session = None) -> Workspace:
        """
        Returns a workspace object for the remote session

        .. highlight:: python
        .. code-block:: python

            from DfiiBridge import VirtuosoSessionManager

            # first create an instance of the SessionManager
            session_manager = VirtuosoSessionManager()

            # Connect to the session of the current project (if available)
            ws = session_manager.get_workspace()

            # Get all sessions manually and select one
            sessions = session_manager.find_sessions(True)
            ws = session_manager.get_workspace(sessions[0])

        """
        # If no session is given try all sessions of the current project
        # Make sure to remove invalid session files
        # this should be done------
        if session is None:
            session = self.find_virtuoso_session_from_env()
            return self.get_workspace(session)

        if isinstance(session, Session):
            return Workspace(session)

        raise Exception(f"Unsupported argument type of {session}")

    def kill_session(self, session: Session):
        """Kill a session safely

        Args:
            session: Session to be killed
        """
        ws = self.get_workspace(session)
        if not ws.execute_raw('DfiiBridge__safeExit( )', request_reply=True, timeout=1000):
            raise Exception("DfiiBridge__safeExit() call returned False. Please report this bug.")

        self.delete_session_file(session)

    def create_virtuoso_workspace(self,
                                  no_cdsinit:bool=False,
                                  no_gui:bool=False,
                                  cmd:Callable=None,
                                  python_executable:str=None,
                                  env:dict=None,
                                  extra_fields:dict=None,
                                  on_output:Callable=None)->Workspace:
        """Create a new Virtuoso session and connect to it

        Args:
            no_cdsinit (bool, optional): Start Virtuoso with the nocdsinit flag. Defaults to False.
            no_gui (bool, optional): Start Virtuoso with the nogui flag. Defaults to False.
            cmd (Callable, optional): Override command to start Virtuoso (lambda returning a list that is passed to Popen()). Defaults to None.
            env (dict, optional): Override the environment to be used. Defaults to None (taking os.environ.copy()).
            extra_fields (dict, optional): Add additional fields to the server. Defaults to None.
            on_output (Callable, optional): Add callback to capture stdout of the subprocess. Defaults to None (printing on stdout).

        Returns:
            Workspace: The workspace object
        """

        proc = SessionMaker.create_session(no_gui=no_gui,
                                           no_cdsinit=no_cdsinit,
                                           cmd=cmd,
                                           env=env,
                                           session_root=self.get_root_directory(),
                                           extra_fields=extra_fields,
                                           python_executable=python_executable)

        if on_output is None:
            on_output_callback = lambda _, msg: print(msg, end="")
        else:
            on_output_callback = on_output

        session = SessionMaker.wait_for_session(proc, self,  on_output=on_output_callback)
        return self.get_workspace(session)

    def find_virtuoso_sessions(self, connection_check:bool=True) -> list[Session]:
        """Find all available virtuoso sessions

        Args:
            connection_check (bool, optional): Check connection before returning. Defaults to True.

        Returns:
            list[Session]: List of valid virtuoso sessions
        """
        search_criteria = {"type": "virtuoso"}
        return self.find_sessions(search_criteria, connection_check)


    def find_virtuoso_session_from_env(self, connection_check:bool=True) -> Session:
        """
        Find a session by reading the $DFIIBRIDGE_VIRTUOSO_SESSION_FILE environment variable.
        Useful, if you start a terminal from within Virtuoso and you want
        to connect to exactly this Virtuoso session.
        """
        session_file = os.getenv("DFIIBRIDGE_VIRTUOSO_SESSION_FILE", default="")

        if session_file:
            logger.info(f"Read session from $DFIIBRIDGE_VIRTUOSO_SESSION_FILE = {session_file}")
            return self.get_session(session_file, connection_check)

        raise Exception(f"The environmentvariable $DFIIBRIDGE_VIRTUOSO_SESSION_FILE not found. Was this shell started from within Virtuoso?")



"""
This module declares types for later usage and annotations

Internal package - does not need to be used
"""

from typing import Union, List, Callable, NewType, NamedTuple, Dict, Set, Tuple
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing_extensions import Protocol
else:
    class Protocol:
        pass


__all__ = [
    'Number',
    'Symbol',
    'Key',
    'Var',
    'SkillComponent',
    'SkillCode',
    'Skill',
    'Replicator',
    'Definition',
    'Function',
    'SkillTuple',
    'SkillList',
]

Number = Union[int, float]
SkillComponent = Union[int, str]
SkillCode = NewType('SkillCode', str)


class Function(NamedTuple):
    name: str
    description: str
    aliases: Set[str]


Definition = List[Function]


class SupportsReprSkill(Protocol):
    def __repr_skill__(self) -> SkillCode:
        ...


Skill = Union[SupportsReprSkill, Number, str, bool, None, 'SkillList', 'SkillDict', 'SkillTuple']
Replicator = Callable[[str], Skill]


class SkillList(List[Skill]):
    pass


class SkillTuple(Tuple[Skill, ...]):
    pass


class SkillDict(Dict[str, Skill]):
    pass


class Symbol(NamedTuple):
    name: str

    def __repr_skill__(self) -> SkillCode:
        return SkillCode(f"'{self.name}")

    def __str__(self) -> str:
        return f"Symbol({self.name})"

    def __repr__(self) -> str:
        return f"Symbol({self.name!r})"


class Key(NamedTuple):
    name: str

    def __repr_skill__(self) -> SkillCode:
        return SkillCode(f"?{self.name}")

    def __str__(self) -> str:
        return f"Key({self.name})"

    def __repr__(self) -> str:
        return f"Key({self.name})"


class Var(NamedTuple):
    name: str

    def __repr_skill__(self) -> SkillCode:
        return SkillCode(self.name)

    def __str__(self) -> str:
        return f"Var({self.name})"

    def __repr__(self) -> str:
        return f"Var({self.name!r})"
